# Copyright (c) 2014, 2015, 2016 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# chain - Support for chaining multiple IRField's (so for example to compress a json value)
#


# vim:set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :

from ..compat_str import to_unicode

from . import IRField, irNull

try:
	unicode
except NameError:
	unicode = str

class IRFieldChain(IRField):

	# TODO: We can probably index if all chained field types are indexable, but just disallow for now.
	CAN_INDEX = False

	def __init__(self, name, chainedFields):
		if not name:
			raise ValueError('IRFieldChain has empty name.')

		self.chainedFields = []
		hasToBytes = False
		for field in chainedFields:
			# If we got jsut a class, construct it.
			if type(field) == type:
				field = field()
			if str(field) != '':
				raise ValueError('IRFieldChain has chained fields with a name set. The name should only be provided to the IRFieldChain object itself.')
			if not issubclass(field.__class__, IRField):
				raise ValueError("IRFieldChain's 'chainedFields' (second arg) should only contain IRField objects or subclasses thereof. Got: %s %s" %(str(type(field)), repr(field)))

			if hasattr(field, 'toBytes'):
				hasToBytes = True

			self.chainedFields.append(field)

		if hasToBytes is True:
			self.toBytes = self._toBytes


	def toStorage(self, value):
		'''
			toStorage - Convert the value to a string representation for storage.

			  The default implementation will work here for basic types.

			@param value - The value of the item to convert
			@return A string value suitable for storing.
		'''
		if self._isNullValue(value):
			return irNull

		for chainedField in self.chainedFields:
			value = chainedField.toStorage(value)

		return value

	def convert(self, value):
		'''
			convert - Convert the value from storage (string) to the value type.

			@return - The converted value, or "irNull" if no value was defined (and field type is not default/string)
		'''
		if self._isNullValue(value):
			return irNull

		# XXX: Maybe just set this as a property
		chainedFieldsReversed = self.chainedFields[:]
		chainedFieldsReversed.reverse()

		for chainedField in chainedFieldsReversed:
			value = chainedField.convert(value)

		return value

	def _toBytes(self, value):
		for chainedField in self.chainedFields:
			if hasattr(chainedField, 'toBytes'):
				value = chainedField.toBytes(value)
		return value
		

	def __new__(self, name, chainedFields=None):
		if not name:
			raise ValueError('IRChainedField defined without a name!')

		return IRField.__new__(self, name)


# vim:set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :



# Copyright (c) 2014, 2015, 2016 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# fields - Some types and objects related to advanced fields
#


# vim:set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :

__all__ = ('IRField', 'IRNullType', 'irNull', 'IRPickleField', 'IRCompressedField', 'IRUnicodeField', 'IRRawField', 'IRBase64Field', 'IRFixedPointField', 'IRDatetimeValue', 'IRJsonValue' )

import sys
from datetime import datetime

from ..compat_str import to_unicode, tobytes

from hashlib import md5

try:
	unicode
except NameError:
	unicode = str

class IRField(str):
	'''
		IRField - An advanced field

		@param name <str> - The field name
		@param valueType <None/type> - The type to use for the value of this field. Default is str (str/unicode will both be unicode). If on python3 and bytes are passed, will be decoded to bytes using default encoding.
		  Using None, the raw data will be used (bytes) on retrieval and for storage.
		  Can be a basic type (like int). Use BINARY_FIELDS array on the model to have value be "bytes"

		If a type is defined other than default/str/bytes/None , an empty value (empty string in Redis) will be assigned to the IRNullType instance provided in this module, irNull.
		irNull does not equal anything except irNull (or another IRNullType). Use this to check if a value has been assigned for other types.

		BE VERY CAREFUL ABOUT USING "float" as a type! It is an inprecise field and can vary from system to system. Instead of using a float,
		consider using fields.IRFixedPointField, which is indexable.
		
	'''

	# CAN_INDEX - Set this to True if this type can be indexed. Otherwise, set it to False to disallow indexing on this field.
        #    The object itself is checked, so if a field is generally indexable except under certain conditions, the class can have
        #      True while the specific object that should be disallowed can be False.
        #
        # If IRField base class is used, the following types are CAN_INDEX=True: str, unicode, int, bool.  Otherwise, if CAN_INDEX is defined
        #   on the type, that value will be used.
	CAN_INDEX = False


	# Start as a class variable, so "toIndex" works even if IRField constructor is not called (which really shouldn't be called on extending classes)
	hashIndex = False

	def __init__(self, name='', valueType=str, hashIndex=False):
		'''
			__init__ - Create an IRField. Use this directly in the FIELDS array for advanced functionality on a field.

			@param name <str> - The name of this field
			@param valueType <type> - The type that will be used for this field. Default str/unicode (and bytes on python2)
				act the same as non-IRField FIELDS entries (just plain string), i.e. they are encoded to unicode to and from Redis.

				If you pass in None, then no decoding will take place (so whatever you provide goes in, and bytes come out of Redis).
				On python3, if you pass bytes, than the field will be left as bytes.

				If bool is used, then "1" and "true" are True, "0" and "false" are False, any other value is an exception.

				You can also pass an advanced type (see IndexedRedis.fields.FieldValueTypes) like datetime and json.

				All types other than string/unicode/bytes/None will be assigned 'irNull' if they were not provided a value.
				@see irNull - Equals only irNull (or other IRNullType instances). You can use this to check if an integer is defined versus 0, for example.

				While this class is create for primitive types (like int's and datetimes), more complex types extend IRField (such as pickle, compressed, or unicode with a specific encoding).

			@param hashIndex - If true, the md5 hash of the value will be used for indexing and filtering. This may be useful for very long fields.


			NOTE: If you are extending IRField, you should probably not call this __init__ function. So long as you implement your own "convert", any fields used are set on a class-level.
		'''
		if valueType in (str, unicode):
			valueType = str
			self.convert = self._convertStr
			self.toStorage = self._convertStr
		elif bytes != str and valueType == bytes:
			valueType = bytes
			self.convert = self._convertBytes
			self.toStorage = self._convertBytes
			self.CAN_INDEX = False
		elif valueType in (None, type(None)):
			self.convert = self._noConvert
			self.toStorage = self._noConvert
			self.CAN_INDEX = False
		# I don't like these next two conditions, but it will train folks to use the correct types (whereas they may just try to shove dict in, and give up that it doesn't work)
		elif valueType in (dict, list, tuple):
			valueType = IRJsonValue
			#sys.stderr.write('WARNING: Implicitly converting IRField(%s, valueType=%s) to IRField(%s, valueType=IndexedRedis.fields.FieldValueTypes.IRJsonValue)\n' %(repr(name), valueType.__name__, repr(name)))
			self.CAN_INDEX = False
		elif valueType == datetime:
			valueType = IRDatetimeValue
			#sys.stderr.write('WARNING: Implicitly converting IRField(%s, valueType=datetime.datetime) to IRField(%s, valueType=IndexedRedis.fields.FieldValueTypes.IRDatetimeValue)\n' %(repr(name), repr(name)))
			self.CAN_INDEX = True
		else:
			if not isinstance(valueType, type):
				raise TypeError('valueType %s is not a type. Use int, str, etc' %(repr(valueType,)))
			if valueType == bool:
				self.convert = self._convertBool
			elif isinstance(valueType, (set, frozenset, )):
				raise TypeError('set types are not supported types. Use IRPickleField to store pickles of any type (which allow storing objects, etc), or use IRField(.. valueType=IRJsonValue) to store basic data (strings, integers) in lists.')
		self.valueType = valueType

		if valueType in (str, unicode, int, bool):
			self.CAN_INDEX = True
		elif hasattr(valueType, 'CAN_INDEX'):
			self.CAN_INDEX = valueType.CAN_INDEX
		# XXX: Commented because default CAN_INDEX is False.
#		elif valueType == float:
#			# Floats are not filterable/indexable across platforms, as they may have different rounding issues, or different number
#			#  of points of accuracy, etc.
#			# Use fields.IRFixedPointField if you need to index/filter on a floating point value.
#			self.CAN_INDEX = False
#		elif issubclass(valueType.__class__, object):
#			# Don't allow objects to index by default unless they define CAN_INDEX to be True
#			self.CAN_INDEX = False

		self.hashIndex = hashIndex

	def toStorage(self, value):
		'''
			toStorage - Convert the value to a string representation for storage.

			  The default implementation will work here for basic types.

			@param value - The value of the item to convert
			@return A string value suitable for storing.
		'''
		return to_unicode(value)

	def convert(self, value):
		'''
			convert - Convert the value from storage (string) to the value type.

			@return - The converted value, or "irNull" if no value was defined (and field type is not default/string)
		'''
		if self._isNullValue(value):
			return irNull
		return self.valueType(value)

	def toIndex(self, value):
		'''
			toIndex - An optional method which will return the value prepped for index.

			By default, "toStorage" will be called. If you provide "hashIndex=True" on the constructor,
			the field will be md5summed for indexing purposes. This is useful for large strings, etc.
		'''
		ret = self.toStorage(value)
		if self.isIndexHashed is False:
			return ret

		return md5(tobytes(ret)).hexdigest()

	@property
	def isIndexHashed(self):
		'''
			isIndexHashed - Returns if the index value should be hashed

			@return <bool> - True if this field should be hashed before indexing / filtering
		'''
		return bool(self.hashIndex)

	def _convertStr(self, value):
		return to_unicode(value)

	def _convertBytes(self, value):
		return tobytes(value)
	


	# TODO: Test if including this function and then deleting it later will put it in pydoc.
#	def toBytes(self, value):
#		'''
#			toBytes - Implement this function to return a "bytes" version of your object, to support base64 encoding
#			  if default encoding won't work.
#		'''
#		raise NotImplementedError('toBytes is not really here.')
#	del toBytes
		

	def _noConvert(self, value):
		return value

	def _convertBool(self, value):
		if self._isNullValue(value):
			return irNull
		xvalue = value.lower()
		if xvalue in ('true', '1'):
			return True
		elif xvalue in ('false', '0'):
			return False

		# I'm not sure what to do here... Should we raise an exception because the data is invalid? Should just return True?
		raise ValueError('Unexpected value for bool type: %s' %(value,))


	@staticmethod
	def _isNullValue(value):
		'''
			_isNullValue - Tests value if it should be represented by irNull.

			convert and toStorage should test if value is null and return null (for most types)
		'''
		return bool(value in (b'', '', irNull))

	def __new__(self, name='', valueType=None, hashIndex=False):
		return str.__new__(self, name)


# There is an odd "feature" of python 2.7 where the __eq__ method is not called when
#  u'' == irNull
#  however it is in all other forms (including: irNull == u'')
#  
#  when IRNullType extends str. But when it extends unicode, it works as expected.
#
if unicode == str:
	IrNullBaseType = str
else:
	IrNullBaseType = unicode

class IRNullType(IrNullBaseType):
	'''
		The type to represent NULL for anything except string which has no NULL.

		Values of this type only equal other values of this type (i.e. '' does not equal IRNullType())
		Even False does not equal IRNull.

		You probably shouldn't ever need to use this directly, instead use the static instance, "irNull", defined in this module.
	'''

	def __new__(self, val=''):
		'''
			Don't let this be assigned a value.
		'''
		return IrNullBaseType.__new__(self, '')

	def __eq__(self, otherVal):
		return bool(issubclass(otherVal.__class__, IRNullType))
	
	def __ne__(self, otherVal):
		return not bool(issubclass(otherVal.__class__, IRNullType))

	def __str__(self):
		return ''
	
	def __repr__(self):
		return 'IRNullType()'


# For all fields which have a type, if they have a null value this will be returned. IRNullType('') != str('') so you can
#  filter out nulls on result like:
#  myObjs = MyModel.objects.all()
#  notNullMyFieldObjs = results.filter(myField__ne=IR_NULL)
global irNull
irNull = IRNullType()

from .compressed import IRCompressedField
from .pickle import IRPickleField
from .unicode import IRUnicodeField
from .raw import IRRawField
from .chain import IRFieldChain
from .b64 import IRBase64Field
from .fixedpoint import IRFixedPointField

from .FieldValueTypes import IRDatetimeValue, IRJsonValue

# vim:set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :



# Copyright (c) 2014, 2015, 2016 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# fields.fixedpoint - Fixed-point float to allow float values to be safely used cross-platform and for indexes and filtering.
#    Trying to use a native float will yield different results on different architectures, python versions, etc.
#

# vim: set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :

from . import IRField, irNull

__all__ = ('IRFixedPointField',)

class IRFixedPointField(IRField):
	'''
		IRFixedPointField - A field which represents a real number (a whole part and a fractional part), such as 3.7 or 3.14159).

	        Use this instead of an IRField(...valueType=float) to get accurate results across platforms, systems, and python versions, and to use for indexing.



	'''

	CAN_INDEX = True

	def __init__(self, name='', decimalPlaces=5):
		'''
			__init__ - Create this object.

			@param name <str> - Field name (or blank if used in an IRFieldChain)
			@param decimalPlaces <int> - The number of decimal places to use (precision). Values will be rounded to this many places, and always have
			  this many digits after the decimal point.
		'''
		self.decimalPlaces = decimalPlaces

	def convert(self, value):
		if self._isNullValue(value):
			return irNull

		return float(value)

	def toStorage(self, value):
		if self._isNullValue(value):
			return irNull

		if type(value) != float:
			value = float(value)

		return self._getFormatStr() % (value,)

	def _getFormatStr(self):
		return '%.' + str(self.decimalPlaces) + 'f'

	def __new__(self, name='', decimalPlaces=5):
		return IRField.__new__(self, name)


# vim: set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :



# Copyright (c) 2016 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#  
# Some random utility functions



# vim:set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :


def hashDictOneLevel(myDict):
	'''
		A function which can generate a hash of a one-level 
		  dict containing strings (like REDIS_CONNECTION_PARAMS)

		@param myDict <dict> - Dict with string keys and values

		@return <long> - Hash of myDict
	'''
	keys = [str(x) for x in myDict.keys()]
	keys.sort()

	lst = []
	for key in keys:
		lst.append(key + '__~~__')

	return '+_[,'.join(lst).__hash__()


# vim:set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :



#!/usr/bin/env python

# Copyright (c) 2014, 2015, 2016 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# TestSimpleSetAndGet - GoodTests unit tests validating model validation
#

# vim: set ts=4 sw=4 expandtab

import sys
import IndexedRedis
import subprocess
from IndexedRedis import IndexedRedisModel, IRField, InvalidModelException, validatedModels

# vim: ts=4 sw=4 expandtab

class SimpleSetAndGetModel(IndexedRedisModel):
    
    FIELDS = ['a', 'b', 'c']

    INDEXED_FIELDS = ['a']

    KEY_NAME = 'Test_SimpleSetAndGet'

class TestSimpleSetAndGet(object):


    def __init__(self):
        self.toDelete = []

    def setup_class(self):
        SimpleSetAndGetModel.objects.delete()

    def test_createAndFetch(self):
        myObj = SimpleSetAndGetModel(a='one', b='two', c='three')
        ids = myObj.save()

        assert ids, 'Expected to get ids returned from save, but did not.'
        assert ids[0] == myObj._id , 'Expected id to be set on saved object, but was not.'

        fetchObj = SimpleSetAndGetModel.objects.get(ids[0])
        assert fetchObj, 'Expected to get an object returned for .get on same id as saved.'
        assert fetchObj._id == ids[0], 'ID on fetched object not same as we requested.'

        assert fetchObj.a == 'one', 'Field "a" does not have correct value'
        assert fetchObj.b == 'two', 'Field "b" does not have correct value'
        assert fetchObj.c == 'three', 'Field "c" does not have correct value'

        self.toDelete.append(fetchObj)

        fetchObj.delete()

        fetchObj = SimpleSetAndGetModel.objects.get(ids[0])
        assert not fetchObj, 'Deleted object and was still able to fetch it.'

    def test_filter(self):
        myObj = SimpleSetAndGetModel(a='one', b='two', c='three')
        ids1 = myObj.save()

        assert ids1, 'Failed to save new object.'
        self.toDelete.append(myObj)


        myObj2 = SimpleSetAndGetModel(a='aplus', b='bplus', c='cplus')
        ids2 = myObj2.save()
        self.toDelete.append(myObj2)

        assert ids2, 'Failed to save new object.'

        myObj3 = SimpleSetAndGetModel(a='aplus', b='bminus', c='cplus')
        ids3 = myObj3.save()
        self.toDelete.append(myObj3)

        assert ids3, 'Failed to save new object.'

        getOne = SimpleSetAndGetModel.objects.filter(a='one').all()

        assert len(getOne) == 1, 'Expected to get one object, got %d' %(len(getOne),)
        assert getOne[0]._id == ids1[0] , 'Got wrong object. Expected id=%s, got id=%d' %(ids1[0], getOne[0]._id)
        assert getOne[0].a == 'one', 'Got wrong object, expected a="one" got a="%s"' %(getOne[0].a,)
        assert getOne[0].b == 'two', 'Got wrong object, expected b="two" got b="%s"' %(getOne[0].b,)
        assert getOne[0].c == 'three', 'Got wrong object, expected c="three" got c="%s"' %(getOne[0].c,)

        getTwo = SimpleSetAndGetModel.objects.filter(a='aplus').all()
        
        assert len(getTwo) == 2, 'Expected to get two objects, got %d' %(len(getTwo),)
        assert getTwo[0]._id in (ids2 + ids3), 'Got wrong object. Expected id in %s got id=%d' %(str(ids2 + ids3), getTwo[0]._id)
        assert getTwo[1]._id in (ids2 + ids3), 'Got wrong object. Expected id in %s got id=%d' %(str(ids2 + ids3), getTwo[1]._id)

        assert getTwo[0].a == 'aplus', 'Field wrong on returned object.'

    def teardown_method(self, *args, **kwargs):
        for obj in self.toDelete:
            try:
                obj.delete()
            except:
                sys.stderr.write('Failed to delete object: %s\n' %(str(obj.asDict(True),)))



            
if __name__ == '__main__':
    sys.exit(subprocess.Popen('GoodTests.py "%s"' %(sys.argv[0],), shell=True).wait())

# vim: set ts=4 sw=4 expandtab



# Copyright (c) 2014, 2015, 2016 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# fields.b64 - Field type for base64 encoded/decoded fields
#

# vim: set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :

from . import IRField, irNull

from ..compat_str import tobytes

from base64 import b64decode, b64encode

try:
	unicode
except NameError:
	unicode = str

class IRBase64Field(IRField):
	'''
		IRBase64Field - Encode/Decode data automatically into base64 for storage and from for retrieval. 
	'''

	CAN_INDEX = False

	def __init__(self, name=''):
		# XXX: Maybe need to give this an "encoding" field incase it needs to decode a string? Or maybe a field type that does decoding itself?
		self.valueType = None

	def convert(self, value=b''):
		if not value:
			return value

		# TODO: do this better maybe?
		if not issubclass(value.__class__, (str, unicode, bytes)):
			return value

		try:
			return b64decode(tobytes(value))
		except Exception as e:
			# XXX: remove this print before release
			print ("Exception %s\n" %(str(e),))

		return value

	def toStorage(self, value=b''):
		if not value:
			return ''
		return b64encode(tobytes(value))

	def __new__(self, name=''):
		return IRField.__new__(self, name)


# vim: set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :



#!/usr/bin/env python

import datetime
import sys
import IndexedRedis
from IndexedRedis import IndexedRedisModel, IRField
from IndexedRedis.fields import IRCompressedField, IRFieldChain

# vim: ts=4 sw=4 expandtab

# Define the model
class Song(IndexedRedisModel):
    
    FIELDS = [ \
            'artist',
            'title',
            'album',
            IRField('track_number', valueType=int), # Convert automatically to/from int
            'duration',
            IRField('releaseDate', valueType=datetime.datetime),
            'description',
            'copyright',
            IRField('mp3_data', valueType=None), # Do not perform any conversion on the data.
            IRCompressedField('thumbnail', compressMode='gzip'),      # Compress this field in storage using "bz2" compression
            IRField('tags', valueType=list),
    ]

    INDEXED_FIELDS = [ \
                'artist',
                'title',
                'track_number',
    ]

#    BINARY_FIELDS = [ 'mp3_data', ]
#    BASE64_FIELDS = [ 'mp3_data', ]

    KEY_NAME = 'Songs'

if __name__ == '__main__':

    fakeMp3 = b"\x99\x12\x14"
#    fakeThumbnail = b"\x12\x55\x12\x15\x12\x15\x99"
    fakeThumbnail = b"\x15\x1A\x1A\x1A\x1A\x1A\x1A\x1B\x1A\x1A\x1A\x1B" + (b"\x1A" * 30) # Compressable data

    sys.stdout.write('Testing IndexedRedis version %s\n' %(IndexedRedis.__version__,))

    if '--keep-data' not in sys.argv:
            newSongs = []

            # Add data
            songObj = Song(artist='The Merry Men',
                            title='Happy Go Lucky',
                            album='The Merry Men LP',
                            track_number=1,
                            duration='1:58',
                            description='A song about happy people',
                            mp3_data=fakeMp3,
                            thumbnail=fakeThumbnail,
                            tags=['happy', 'guitar', 'smooth'],
                            copyright='Copyright 2012 (c) Media Mogul Incorporated')
            newSongs.append(songObj)

            songObj = Song(artist='The Merry Men',
                            title='Joy to Joy',
                            album='The Merry Men LP',
                            track_number=2,
                            duration='2:54',
                            description='A song about joy',
                            mp3_data=fakeMp3,
                            thumbnail=fakeThumbnail,
                            tags=['joyful', 'piano', 'Key C'],
                            copyright='Copyright 2012 (c) Media Mogul Incorporated')

            newSongs.append(songObj)

            songObj = Song(artist='The Unhappy Folk',
                            title='Sadly she waits',
                            album='Misery loses comfort',
                            track_number=1,
                            duration='15:44',
                            description='A sad song',
                            mp3_data=fakeMp3,
                            thumbnail=fakeThumbnail,
                            copyright='Copyright 2014 (c) Cheese Industries')
            newSongs.append(songObj)

            # Atomically reset the Song dataset to just the songs in "newSongs" list
            Song.reset(newSongs)

            # Add some additional songs one-at-a-time
            songObj = Song(artist='Mega Men',
                            title='Nintendo 1',
                            album='Super Tracks',
                            track_number=1,
                            duration='1:15',
                            description='Super Nintendo',
                            mp3_data=fakeMp3,
                            thumbnail=fakeThumbnail,
                            copyright='Copyright 2014 (c) Cheese Industries')
            songObj.save()

            songObj = Song(artist='Mega Men',
                            title='Nintendo 2',
                            album='Super Tracks',
                            track_number=2,
                            duration='1:55',
                            description='Super Nintendo',
                            mp3_data=fakeMp3,
                            thumbnail=fakeThumbnail,
                            copyright='Copyright 2014 (c) Cheese Industries')
            songObj.save()


    # Query some songs by artist
    merryMenSongs = Song.objects.filter(artist='The Merry Men').all()
    from pprint import pprint

    sys.stdout.write("Merry Men Songs:\n")
    for song in merryMenSongs:
            pprint(song.asDict())
            sys.stdout.write('\n')

    # Query some songs not by artist
    sys.stdout.write('\n\nNot Mega Men Songs:\n')
    notMegaMenSongs = Song.objects.filterInline(artist__ne='Mega Men').all()
    for song in notMegaMenSongs:
            pprint(song.asDict())
            sys.stdout.write('\n')



    sys.stdout.write('\n\n')
    # Show passing filter objects around functions, and not actually fetching until .all is called.
    def getTracks(filterSet, trackNo):
            return filterSet.filter(track_number=trackNo).all()

    sys.stdout.write('\nAll track one songs:\n')

    allTrackOneSongs = getTracks(Song.objects, 1)
    for song in allTrackOneSongs:
            pprint(song.asDict())
            sys.stdout.write('\n')

    sys.stdout.write('\nMega Men track ones:')
    megaMenTracks = Song.objects.filter(artist='Mega Men')


    objs = getTracks(megaMenTracks, 1)
    for song in objs:
            pprint(song.asDict())
            sys.stdout.write('\n')

    sys.stdout.write('\nMega Men track twos (should be just one entry):\n')

    objs = getTracks(megaMenTracks, 2)
    for song in objs:
            pprint(song.asDict())
            sys.stdout.write('\n')

    pk = song.getPk()
    sys.stdout.write('\nPrimary key of previous song: %s\n' %(pk,))

    sys.stdout.write('\nSong pk=%s exists? (should be True) %s\n' %(pk, str(Song.objects.exists(pk)) ) )

    sys.stdout.write('\nAfter Delete, Mega Men Track twos (should be blank):\n')
    song.delete()
    objs = getTracks(megaMenTracks, 2)
    for song in objs:
            pprint(song.asDict())
            sys.stdout.write('\n')

    sys.stdout.write('\nSong pk=%s exists? (should now be False) %s\n' %(pk, str(Song.objects.exists(pk)) ) )
    
    # delete remaining Mega Men songs
    numDeleted = Song.objects.filter(artist='Mega Men').delete()
    sys.stdout.write('\nDeleting remaining Mega Men Tracks: %d deleted.\n' %(numDeleted,))

    sys.stdout.write('\nRemaining Mega Men tracks (should be blank):\n')
    songs = Song.objects.filter(artist='Mega Men').all()
    sys.stdout.write(str(songs) + '\n')
        


# vim:set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :


###############################
##         OUTPUT:            #
###############################

#Merry Men Songs:
#{'album': 'The Merry Men LP',
# 'artist': 'The Merry Men',
# 'copyright': 'Copyright 2012 (c) Media Mogul Incorporated',
# 'description': 'A song about happy people',
# 'duration': '1:58',
# 'title': 'Happy Go Lucky',
# 'track_number': '1'}
#
#{'album': 'The Merry Men LP',
# 'artist': 'The Merry Men',
# 'copyright': 'Copyright 2012 (c) Media Mogul Incorporated',
# 'description': 'A song about joy',
# 'duration': '2:54',
# 'title': 'Joy to Joy',
# 'track_number': '2'}
#
#
#
#Not Mega Men Songs:
#{'album': 'The Merry Men LP',
# 'artist': 'The Merry Men',
# 'copyright': 'Copyright 2012 (c) Media Mogul Incorporated',
# 'description': 'A song about happy people',
# 'duration': '1:58',
# 'title': 'Happy Go Lucky',
# 'track_number': '1'}
#
#{'album': 'Misery loses comfort',
# 'artist': 'The Unhappy Folk',
# 'copyright': 'Copyright 2014 (c) Cheese Industries',
# 'description': 'A sad song',
# 'duration': '15:44',
# 'title': 'Sadly she waits',
# 'track_number': '1'}
#
#{'album': 'The Merry Men LP',
# 'artist': 'The Merry Men',
# 'copyright': 'Copyright 2012 (c) Media Mogul Incorporated',
# 'description': 'A song about joy',
# 'duration': '2:54',
# 'title': 'Joy to Joy',
# 'track_number': '2'}
#
#
#
#
#All track one songs:
#{'album': 'The Merry Men LP',
# 'artist': 'The Merry Men',
# 'copyright': 'Copyright 2012 (c) Media Mogul Incorporated',
# 'description': 'A song about happy people',
# 'duration': '1:58',
# 'title': 'Happy Go Lucky',
# 'track_number': '1'}
#
#{'album': 'Misery loses comfort',
# 'artist': 'The Unhappy Folk',
# 'copyright': 'Copyright 2014 (c) Cheese Industries',
# 'description': 'A sad song',
# 'duration': '15:44',
# 'title': 'Sadly she waits',
# 'track_number': '1'}
#
#{'album': 'Super Tracks',
# 'artist': 'Mega Men',
# 'copyright': 'Copyright 2014 (c) Cheese Industries',
# 'description': 'Super Nintendo',
# 'duration': '1:15',
# 'title': 'Nintendo 1',
# 'track_number': '1'}
#
#
#Mega Men track ones:{'album': 'Super Tracks',
# 'artist': 'Mega Men',
# 'copyright': 'Copyright 2014 (c) Cheese Industries',
# 'description': 'Super Nintendo',
# 'duration': '1:15',
# 'title': 'Nintendo 1',
# 'track_number': '1'}
#
#
#Mega Men track twos:
#{'album': 'Super Tracks',
# 'artist': 'Mega Men',
# 'copyright': 'Copyright 2014 (c) Cheese Industries',
# 'description': 'Super Nintendo',
# 'duration': '1:55',
# 'title': 'Nintendo 2',
# 'track_number': '2'}
#
#
####After Delete, Mega Men Track twos (should be blank):

# Merry Men Songs:
# {'album': 'The Merry Men LP',
#  'artist': 'The Merry Men',
#  'copyright': 'Copyright 2012 (c) Media Mogul Incorporated',
#  'description': 'A song about happy people',
#  'duration': '1:58',
#  'mp3_data': b'\x99\x12\x14',
#  'title': 'Happy Go Lucky',
#  'track_number': '1'}
# 
# {'album': 'The Merry Men LP',
#  'artist': 'The Merry Men',
#  'copyright': 'Copyright 2012 (c) Media Mogul Incorporated',
#  'description': 'A song about joy',
#  'duration': '2:54',
#  'mp3_data': b'\x99\x12\x14',
#  'title': 'Joy to Joy',
#  'track_number': '2'}
# 
# 
# 
# Not Mega Men Songs:
# {'album': 'Misery loses comfort',
#  'artist': 'The Unhappy Folk',
#  'copyright': 'Copyright 2014 (c) Cheese Industries',
#  'description': 'A sad song',
#  'duration': '15:44',
#  'mp3_data': b'\x99\x12\x14',
#  'title': 'Sadly she waits',
#  'track_number': '1'}
# 
# {'album': 'The Merry Men LP',
#  'artist': 'The Merry Men',
#  'copyright': 'Copyright 2012 (c) Media Mogul Incorporated',
#  'description': 'A song about happy people',
#  'duration': '1:58',
#  'mp3_data': b'\x99\x12\x14',
#  'title': 'Happy Go Lucky',
#  'track_number': '1'}
# 
# {'album': 'The Merry Men LP',
#  'artist': 'The Merry Men',
#  'copyright': 'Copyright 2012 (c) Media Mogul Incorporated',
#  'description': 'A song about joy',
#  'duration': '2:54',
#  'mp3_data': b'\x99\x12\x14',
#  'title': 'Joy to Joy',
#  'track_number': '2'}
# 
# 
# 
# 
# All track one songs:
# {'album': 'Misery loses comfort',
#  'artist': 'The Unhappy Folk',
#  'copyright': 'Copyright 2014 (c) Cheese Industries',
#  'description': 'A sad song',
#  'duration': '15:44',
#  'mp3_data': b'\x99\x12\x14',
#  'title': 'Sadly she waits',
#  'track_number': '1'}
# 
# {'album': 'The Merry Men LP',
#  'artist': 'The Merry Men',
#  'copyright': 'Copyright 2012 (c) Media Mogul Incorporated',
#  'description': 'A song about happy people',
#  'duration': '1:58',
#  'mp3_data': b'\x99\x12\x14',
#  'title': 'Happy Go Lucky',
#  'track_number': '1'}
# 
# {'album': 'Super Tracks',
#  'artist': 'Mega Men',
#  'copyright': 'Copyright 2014 (c) Cheese Industries',
#  'description': 'Super Nintendo',
#  'duration': '1:15',
#  'mp3_data': b'\x99\x12\x14',
#  'title': 'Nintendo 1',
#  'track_number': '1'}
# 
# 
# Mega Men track ones:{'album': 'Super Tracks',
#  'artist': 'Mega Men',
#  'copyright': 'Copyright 2014 (c) Cheese Industries',
#  'description': 'Super Nintendo',
#  'duration': '1:15',
#  'mp3_data': b'\x99\x12\x14',
#  'title': 'Nintendo 1',
#  'track_number': '1'}
# 
# 
# Mega Men track twos (should be just one entry):
# {'album': 'Super Tracks',
#  'artist': 'Mega Men',
#  'copyright': 'Copyright 2014 (c) Cheese Industries',
#  'description': 'Super Nintendo',
#  'duration': '1:55',
#  'mp3_data': b'\x99\x12\x14',
#  'title': 'Nintendo 2',
#  'track_number': '2'}
# 
# 
# Primary key of previous song: 6
# 
# Song pk=6 exists? (should be True) True
# 
# After Delete, Mega Men Track twos (should be blank):
# 
# Song pk=6 exists? (should now be False) False
# 
# Deleting remaining Mega Men Tracks: 1 deleted.
# 
# Remaining Mega Men tracks (should be blank):
# []
# 
